/* testlibm - test the system math library
 *
 * Link with -lm
 *
 * Usage: testlibm [ -xml ]
 * Standard input: results from "readres -f cr" or "t3-lastep -m"
 * Standard output: errors
 *
 * Copyright (c) 2003-2015 Vincent Lefevre <vincent@vinc17.net>
 * AriC project, INRIA / LIP / ENS-Lyon, France.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110-1301, USA.
 */

#define SVNID "$Id: testlibm.c 85213 2015-12-24 17:37:15Z vinc17/zira $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <errno.h>
#include <time.h>

#if NO_PROGRESS

#define PROGRESS_INIT ((void) 0)
#define PROGRESS 0

#elif POSIX || __GLIBC__ || __linux__

#include <unistd.h>
#include <sys/times.h>

static struct tms tms_buf;
static unsigned long time_inc = 0;
static unsigned long clk;

static void progress_init(void)
{
  if (isatty (2))
    {
      time_inc = sysconf(_SC_CLK_TCK) / 2;  /* progress info every 1/2 s */
      clk = times(&tms_buf);
    }
}

static int progress(void)
{
  unsigned long new_clk;

  new_clk = times(&tms_buf);
  return new_clk >= clk ? (clk = new_clk + time_inc, 1) : 0;
}

#define PROGRESS_INIT (progress_init())
#define PROGRESS (time_inc && progress())

#else

#define PROGRESS_INIT ((void) 0)
#define PROGRESS (line % 1000 == 0)

#endif

#ifdef LIBMCR
#include <fenv.h>
#include <mcr.h>
#else
#include <math.h>
#endif

#ifdef HAVE_SUNMATH_H
# include <sunmath.h>
# ifndef HAVE_EXPM1
#  define HAVE_EXPM1
# endif
# ifndef HAVE_EXP2
#  define HAVE_EXP2
# endif
# ifndef HAVE_EXP10
#  define HAVE_EXP10
# endif
# ifndef HAVE_CBRT
#  define HAVE_CBRT
# endif
#endif

#if __STDC_VERSION__ >= 199901L
# ifndef HAVE_EXPM1
#  define HAVE_EXPM1
# endif
# ifndef HAVE_EXP2
#  define HAVE_EXP2
# endif
# ifndef HAVE_CBRT
#  define HAVE_CBRT
# endif
#endif

typedef struct
{
  char sname[4];
  char lname[8];
  double (*f)(double);
  unsigned long n;
  unsigned long fail;
} Function;

#ifdef HAVE_EXP10
double exp10(double);
#endif

static double isq(double x)
{
  return pow(x, -2.0);
}

static double isr(double x)
{
  return pow(x, -0.5);
}

static double cube(double x)
{
  return pow(x, 3.0);
}

#ifdef LIBMCR

static double pexp2(double x)
{
  return pow(2.0, x);
}

static double pexp10(double x)
{
  return pow(10.0, x);
}

#endif

static Function fn[] = {
  { "exp", "exp",   exp   },
  { "log", "log",   log   },
#ifndef LIBMCR
#ifdef HAVE_EXPM1
  { "em1", "expm1", expm1 },
  { "l1p", "log1p", log1p },
#endif
#ifdef HAVE_EXP2
  { "ex2", "exp2",  exp2  },
  { "lg2", "log2",  log2  },
#endif
#ifdef HAVE_EXP10
  { "e10", "exp10", exp10 },
#endif
  { "l10", "log10", log10 },
  { "sh",  "sinh",  sinh  },
  { "ash", "asinh", asinh },
  { "ch",  "cosh",  cosh  },
  { "ach", "acosh", acosh },
  { "th",  "tanh",  tanh  },
  { "ath", "atanh", atanh },
#endif
  { "sin", "sin",   sin   },
#ifndef LIBMCR
  { "asn", "asin",  asin  },
#endif
  { "cos", "cos",   cos   },
#ifndef LIBMCR
  { "acs", "acos",  acos  },
#endif
  { "tan", "tan",   tan   },
  { "atn", "atan",  atan  },
  { "isq", "isq",   isq   },
  { "isr", "isqrt", isr   },
  { "cub", "cube",  cube  },
#ifndef LIBMCR
#ifdef HAVE_CBRT
  { "cbr", "cbrt",  cbrt  },
#endif
#endif
#ifdef LIBMCR
  { "ex2", "pexp2",  pexp2  },
  { "e10", "pexp10", pexp10 },
#endif
};

#define NFUNC (sizeof(fn) / sizeof(Function))

static double bintod(char *s)
{
  double r, t;
  char *end, *p = s;
  char c;
  long e;

  r = t = s[0] != '-' ? 1.0 : (p++, -1.0);
  if (*p++ != '1' || *p++ != '.')
    {
      fprintf(stderr, "\ntestlibm: bad format in bintod(\"%s\")\n", s);
      exit(3);
    }
  while ((c = *p++) != 'E')
    {
      t /= 2.0;
      if (c == '1')
        r += t;
      else if (c != '0')
        {
          fprintf(stderr, "\ntestlibm: bad format in bintod(\"%s\")\n", s);
          exit(4);
        }
    }
  /* Set errno to 0 since it will be tested just after strtol.
     In particular, errno can be non-zero if ldexp overflowed
     in the previous call to bintod. */
  errno = 0;
  e = strtol(p, &end, 10);
  if (*end != '\0' || errno == ERANGE || e > INT_MAX || e < INT_MIN)
    {
      fprintf(stderr, "\ntestlibm: bad format in bintod(\"%s\")\n", s);
      exit(5);
    }
  return e < -1022 ? 0.0 : ldexp(r, e);
}

int main(int argc, char **argv)
{
  unsigned long line = 0;
  double x, y, z;
  char rm[2], f[6], bx[256], by[256];
  int xml = 0, err = 0;
  int i;
  clock_t t;

#define VERSION_INFO_STREAM stderr
#include "version-info.h"

  for (i = 1; i < argc; i++)
    {
      if (strcmp(argv[i], "-xml") == 0)
        xml = 1;
    }

  PROGRESS_INIT;
  t = clock();

  while (scanf("%1s %5s %255s %255s %*[01]\n", rm, f, bx, by) == 4)
    {
      char *px, *py;

      line++;
      if (PROGRESS)
        {
          fflush(stdout);
          fprintf(stderr, "Line %lu\r", line);
        }

      if ((bx[0] != '1' && bx[0] != '-') || (by[0] != '1' && by[0] != '-'))
        {
          fprintf(stderr, "\ntestlibm: bad format (line %lu)\n", line);
          exit(1);
        }

      px = strchr(bx, 'E');
      py = strchr(by, 'E');
      if (px == NULL || py == NULL)
        {
          fprintf(stderr, "\ntestlibm: bad format (line %lu)\n", line);
          exit(2);
        }
      if (px - bx - (bx[0] == '-' ? 2 : 1) != 53 ||
          py - by - (by[0] == '-' ? 2 : 1) != 53)
        continue;  /* only test double precision */

#ifdef LIBMCR

      switch (rm[0])
        {

#ifdef FE_TONEAREST
        case 'N':
          if (fesetround (FE_TONEAREST))
            {
              fprintf(stderr, "\ntestlibm: can't set rounding mode to "
                      "FE_TONEAREST\n");
              exit(6);
            }
          break;
#endif

#ifdef FE_TOWARDZERO
        case 'Z':
          if (fesetround (FE_TOWARDZERO))
            {
              fprintf(stderr, "\ntestlibm: can't set rounding mode to "
                      "FE_TOWARDZERO\n");
              exit(6);
            }
          break;
#endif

        default:
          continue;

        }

#else

      if (rm[0] != 'N')
        continue;  /* only test rounding to the nearest mode with libm */

#endif

      x = bintod(bx);
      y = bintod(by);

      for (i = 0; i < NFUNC; i++)
        if (strcmp(f, fn[i].sname) == 0 && x != 0 && y != 0)
          {
            z = (fn[i].f)(x);
            fn[i].n++;
            if (y != z)
              {
                fn[i].fail++;
                if (!xml)
                  {
                    printf("Error: %s[%c] (line %lu)\n", f, rm[0], line);
                    printf("  [arg.]  x = %a (%.17g)\n", x, x);
                    printf("  [corr]  y = %a (%.17g)\n", y, y);
                    printf("  [libm]  z = %a (%.17g)\n", z, z);
                    err = 1;
                  }
              }
          } /* end if */
    } /* end while */

  t = clock() - t;

  /* Note: Platforms have a centisecond accuracy, in general.
     Therefore, no need for more precision. */
  if (err)
    putchar('\n');
  printf("Tested worst cases (in %.2f seconds):\n",
         (double) t / CLOCKS_PER_SEC);
  if (xml)
    puts("  <results>");
  for (i = 0; i < NFUNC; i++)
    printf((xml ?
            "    <fct name=\"%s\" tested=\"%lu\" failed=\"%lu\"/>\n" :
            "  %-6s %6lu tested, %5lu failed\n"),
           fn[i].lname, fn[i].n, fn[i].fail);
  if (xml)
    puts("  </results>");

  return 0;
}
