/* testmpfr - test the GNU MPFR library
 *
 * Link with -lmpfr -lgmp -lm *in that order*
 *
 * Standard input: results from "readres -f cr" or "t3-lastep -m"
 * Standard output: errors
 * Exit status:
 *   0: no errors.
 *   1: the test completed, with wrong ternary values only.
 *   2: the test completed, with wrong results only.
 *   3: the test completed, with both wrong results and ternary values.
 *   4: internal error.
 *   5: bad file format.
 *
 * Copyright (c) 2003-2015 Vincent Lefevre <vincent@vinc17.net>
 * AriC project, INRIA / LIP / ENS-Lyon, France.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110-1301, USA.
 */

#define SVNID "$Id: testmpfr.c 85213 2015-12-24 17:37:15Z vinc17/zira $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gmp.h>
#include <mpfr.h>

#if NO_PROGRESS

#define PROGRESS_INIT ((void) 0)
#define PROGRESS 0

#elif POSIX || __GLIBC__ || __linux__

#include <unistd.h>
#include <sys/times.h>

static struct tms tms_buf;
static unsigned long time_inc = 0;
static unsigned long clk;

static void progress_init (void)
{
  if (isatty (2))
    {
      time_inc = sysconf (_SC_CLK_TCK) / 2;  /* progress info every 1/2 s */
      clk = times (&tms_buf);
    }
}

static int progress (void)
{
  unsigned long new_clk;

  new_clk = times (&tms_buf);
  return new_clk >= clk ? (clk = new_clk + time_inc, 1) : 0;
}

#define PROGRESS_INIT (progress_init ())
#define PROGRESS (time_inc && progress ())

#else

#define PROGRESS_INIT ((void) 0)
#define PROGRESS (line % 1000 == 0)

#endif

typedef struct
{
  char sname[4];
  char lname[8];
  int (*f) ();
  unsigned long n;
  unsigned long fail;
  unsigned long wtv;
} Function;

#define POW(fct)                                                \
static mpfr_t fct ## exp;                                       \
static int fct (mpfr_ptr z, mpfr_srcptr x, mp_rnd_t rnd_mode)   \
{                                                               \
  return mpfr_pow (z, x, fct ## exp, rnd_mode);                 \
}

POW(isq)
POW(isr)
POW(cub)

static Function fn[] = {
  { "exp", "exp",   mpfr_exp   },
  { "log", "log",   mpfr_log   },
  { "em1", "expm1", mpfr_expm1 },
  { "l1p", "log1p", mpfr_log1p },
  { "ex2", "exp2",  mpfr_exp2  },
  { "lg2", "log2",  mpfr_log2  },
  { "e10", "exp10", mpfr_exp10 },
  { "l10", "log10", mpfr_log10 },
  { "sh",  "sinh",  mpfr_sinh  },
  { "ash", "asinh", mpfr_asinh },
  { "ch",  "cosh",  mpfr_cosh  },
  { "ach", "acosh", mpfr_acosh },
  { "th",  "tanh",  mpfr_tanh  },
  { "ath", "atanh", mpfr_atanh },
  { "sin", "sin",   mpfr_sin   },
  { "asn", "asin",  mpfr_asin  },
  { "cos", "cos",   mpfr_cos   },
  { "acs", "acos",  mpfr_acos  },
  { "tan", "tan",   mpfr_tan   },
  { "atn", "atan",  mpfr_atan  },
  { "isq", "isq",   isq        },
  { "isr", "isqrt", isr        },
#if MPFR_VERSION >= MPFR_VERSION_NUM(2,4,0)
  { "isr", "rsqrt", mpfr_rec_sqrt },
#endif
  { "cub", "cube",  cub        },
  { "cbr", "cbrt",  mpfr_cbrt  },
  { "p", "x^n",    mpfr_pow_ui },
  { "r", "x^(1/n)", mpfr_root  },
};

static unsigned long n;  /* 0, else n if function is x^n or x^(1/n). */

#define NFUNC (sizeof(fn) / sizeof(Function))

static int test (mp_rnd_t rnd, int i, mpfr_t x[], int rb, int neg)
{
  int tern;
  int fail = 0;
  int k = 1;

  if (n == 0)
    tern = (fn[i].f) (x[2], x[0], rnd);
  else
    tern = (fn[i].f) (x[2], x[0], n, rnd);

  switch (rnd)
    {
    case MPFR_RNDN:
      if ((rb == neg && tern >= 0) || (rb != neg && tern <= 0))
        fail |= 1;
      break;
    case MPFR_RNDZ:
      if ((!neg && tern >= 0) || (neg && tern <= 0))
        fail |= 1;
      break;
    case MPFR_RNDD:
      if (tern >= 0)
        fail |= 1;
      if (neg)
        {
          k = 3;
          mpfr_set (x[3], x[1], rnd);
          mpfr_nextbelow (x[3]);
        }
      break;
    case MPFR_RNDU:
      if (tern <= 0)
        fail |= 1;
      if (!neg)
        {
          k = 3;
          mpfr_set (x[3], x[1], rnd);
          mpfr_nextabove (x[3]);
        }
      break;
    default:
      fprintf (stderr, "testmpfr: internal error\n");
      exit (4);
    }

  if (mpfr_cmp (x[2], x[k]) != 0)
    fail |= 2;

  if (fail)
    {
      printf ("Error: %s, rounding mode %s\n",
              fn[i].sname, mpfr_print_rnd_mode (rnd));
      printf ("     x = ");
      mpfr_out_str (stdout, 2, 0, x[0], MPFR_RNDN);
      printf ("\n");
      printf ("  corr = ");
      mpfr_out_str (stdout, 2, 0, x[k], MPFR_RNDN);
      printf ("\n");
      printf ("  mpfr = ");
      mpfr_out_str (stdout, 2, 0, x[2], MPFR_RNDN);
      printf ("\n     t = %d\n", tern);
    }

  return fail;
}

int main (void)
{
  clock_t t;
  mpfr_t x[4];
  unsigned long line = 0;
  char rm[2], f[6], b[2][256];
  int rb;  /* rounding bit */
  int i;
  int err = 0;

#define VERSION_INFO_STREAM stderr
#include "version-info.h"

  mpfr_init2 (isqexp, 2);
  mpfr_init2 (isrexp, 2);
  mpfr_init2 (cubexp, 2);
  mpfr_set_d (isqexp, -2.0, MPFR_RNDN);
  mpfr_set_d (isrexp, -0.5, MPFR_RNDN);
  mpfr_set_d (cubexp, 3.0, MPFR_RNDN);

  PROGRESS_INIT;
  t = clock ();

  for (i = 0; i < 4; i++)
    mpfr_init (x[i]);

  while (scanf ("%1s %5s %255s %255s %1d\n", rm, f, b[0], b[1], &rb) == 5)
    {
      char *p[2];
      mp_prec_t prec;

      line++;
      if (PROGRESS)
        {
          fflush (stdout);
          fprintf (stderr, "Line %lu\r", line);
        }

      if ((b[0][0] != '1' && b[0][0] != '-') ||
          (b[1][0] != '1' && b[1][0] != '-') ||
          rb < 0 || rb > 1)
        {
          fprintf (stderr, "testmpfr: bad format (line %lu)\n", line);
          exit (5);
        }

      for (i = 0; i < 2; i++)
        {
          mp_prec_t preci;

          p[i] = strchr (b[i], 'E');
          if (p[i] == NULL)
            {
              fprintf (stderr, "testmpfr: bad format (line %lu)\n", line);
              exit (5);
            }
          preci = p[i] - b[i] - (b[i][0] == '-' ? 2 : 1);
          if (i == 0)
            prec = preci;
          else if (prec != preci)
            {
              fprintf (stderr,
                       "testmpfr: precisions are different (line %lu)\n",
                       line);
              exit (5);
            }
        }

      for (i = 0; i < 4; i++)
        mpfr_set_prec (x[i], prec);

      for (i = 0; i < 2; i++)
        mpfr_set_str (x[i], b[i], 2, MPFR_RNDZ);

      n = 0;
      for (i = 0; i < NFUNC; i++)
        if (strlen (fn[i].sname) > 1 ? strcmp (f, fn[i].sname) == 0 :
            (f[0] == fn[i].sname[0] && (n = strtoul (f+1, NULL, 10)) > 0))
          {
            int fail;

            switch (rm[0])
              {
#define TEST(RND) test (RND, i, x, rb, b[1][0] == '-')
              case 'N':
                fail = TEST (MPFR_RNDN);
                break;
              case 'Z':
                fail = TEST (MPFR_RNDZ) | TEST (MPFR_RNDD) | TEST (MPFR_RNDU);
                break;
              default:
                fprintf (stderr, "testmpfr: bad rounding mode (line %lu)\n",
                         line);
                exit (5);
              }
            err |= fail;
            fn[i].n++;
            fn[i].fail += fail >> 1;
            fn[i].wtv += fail & 1;
          } /* end if */
    } /* end while */

  for (i = 0; i < 4; i++)
    mpfr_clear (x[i]);

  if (err)
    putchar ('\n');
  printf ("Tested worst cases (in %.2f seconds):\n",
          (double) (clock () - t) / CLOCKS_PER_SEC);
  for (i = 0; i < NFUNC; i++)
    printf ("  %-7s %6lu tested, %5lu failed, %5lu with wrong ternary value\n",
            fn[i].lname, fn[i].n, fn[i].fail, fn[i].wtv);

  return err;
}
